#include "event_system.h"
#include "aliases.h"
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <sys/mman.h>

#define INVALID_SYSTEM ((EventSystem){ .id = 0 })
#define INVALID_EVENT ((Event) { .id = 0 })
#define INVALID_LISTENER ((EventListener){ .event = INVALID_EVENT, .id = 0 })

#define IS_INVALID_SYSTEM(ES) (ES.id == 0)
#define IS_INVALID_EVENT(EV) (EV.id == 0)
#define IS_INVALID_LISTENER(LS) (LS.id == 0 || IS_INVALID_EVENT(LS.event))

typedef struct es_internal ESInternal;
struct es_internal {
  u64  alloc_size;
  u64  event_capacity;
  u64  event_count;
  u64  free_count;
  u64  initial_callback_capacity;
  uptr events_offset;
  uptr free_offset;
  uptr callbacks_offset;
};

typedef struct event_slot EventSlot;
struct event_slot {
  uptr offset;
  bool registered;
};

typedef struct event_internal EventInternal;
struct event_internal {
  u64           capacity;
  u64           count;
  u64           free_count; 
  u64           *free;
  EventCallback *callbacks;
};

internal void          *alloc(u64 size);
internal ESInternal    *get_event_system_internal(EventSystem event_system);
internal EventSlot     *get_event_slots_array(const ESInternal *es);
internal u64           *get_free_array(const ESInternal *es);
internal EventInternal *get_event(const ESInternal *es, Event event);
internal EventInternal *get_event_from_offset(const ESInternal *es, u64 offset);

EventSystem es_init(u64 initial_event_capacity, u64 initial_listeners_capacity) {
  // Double event capacity to reduce chances of reallocation
  u64 event_capacity     = initial_event_capacity * 2;
  // Double listener capacity to reduce chances of reallocation
  u64 callback_capacity  = initial_listeners_capacity * 2;
  u64 system             = sizeof(ESInternal);
  // Increase event capacity by one to account for 0th Event being invalid
  u64 event_offsets      = (event_capacity + 1) * sizeof(EventSlot);
  u64 free_slots         = event_capacity * sizeof(u64);
  u64 events             = event_capacity * sizeof(EventInternal);
  // Increase listener capacity by one to account for 0th EventListener being invalid
  u64 event_callbacks    = (callback_capacity + 1) * sizeof(EventCallback *);
  u64 event_free         = callback_capacity * sizeof(u64 *);
  u64 callbacks          = event_capacity * event_callbacks;
  u64 free_callbacks     = event_capacity * event_free;
  u64 alloc_size         = system + event_offsets + free_slots + events + callbacks + free_callbacks;

  EventSystem es = INVALID_SYSTEM;

  // NOTE (Abdelrahman): The idea here is to have a single buffer that holds all the data.
  // For the purposes of this test, offsets are manually calculated to different parts of the buffer.
  // An actual implementation of this should probably use an internal memory arena making allocations
  // easier, while still maintaining all data in a single buffer.
  u8 *buffer = alloc(alloc_size);
  if (!buffer) {
    goto RETURN_SYSTEM;
  }

  memset(buffer, 0, alloc_size);

  es.id = (u64)buffer;

  ESInternal *event_system = (ESInternal *)buffer;

  event_system->alloc_size                = alloc_size;
  event_system->event_capacity            = event_capacity;
  event_system->event_count               = 0;
  event_system->free_count                = 0;
  event_system->initial_callback_capacity = callback_capacity;
  event_system->events_offset             = (uptr)system;
  event_system->free_offset               = (uptr)(system + event_offsets);
  event_system->callbacks_offset          = (uptr)(system + event_offsets + free_slots);

  EventSlot *ev_slots = get_event_slots_array(event_system);

  u64 event_item_capacity = sizeof(EventInternal) + event_callbacks + event_free;
  EventInternal *event;
  uptr offset;
  for (u64 i = 0; i < event_capacity; ++i) {
    offset                   = event_system->callbacks_offset + i * event_item_capacity;
    ev_slots[i + 1].offset   = offset;
    event                    = get_event_from_offset(event_system, offset);
    event->capacity          = callback_capacity;
    // WARN (Abdelrahman): As mentioned above, an actual arena would be very useful here. The below
    // would break if the order of the EventInternal struct changes
    event->free              = (u64 *)((uptr)event_system + offset + sizeof(EventInternal));
    event->callbacks         = (EventCallback *)((uptr)event_system + offset + sizeof(EventInternal) + event_free);
  }

RETURN_SYSTEM:
  return es;
}

void es_deinit(EventSystem *event_system) {
  if (!event_system || IS_INVALID_SYSTEM((*event_system))) {
    return;
  }

  ESInternal *es = get_event_system_internal(*event_system);

  munmap(system, es->alloc_size);

  event_system->id = 0;
}

Event es_register_event(EventSystem event_system) {
  Event event = INVALID_EVENT;

  if (IS_INVALID_SYSTEM(event_system)) {
    goto RETURN_EVENT;
  }

  ESInternal *es                = get_event_system_internal(event_system);
  EventSlot *event_slots_array  = get_event_slots_array(es);

  if (es->free_count > 0) {
    u64 *free_array = get_free_array(es);

    event.id = free_array[--(es->free_count)];
    event_slots_array[event.id].registered = true;

    free_array[es->free_count] = 0;

    goto RETURN_EVENT;
  }

  if (es->event_count >= es->event_capacity) {
    // TODO (Abdelrahman): Handle reallocating when out of events
    goto RETURN_EVENT;
  }

  event.id = ++(es->event_count);
  event_slots_array[event.id].registered = true;

RETURN_EVENT:
  return event;
}

void es_deregister_event(EventSystem event_system, Event event) {
  if (IS_INVALID_SYSTEM(event_system) || IS_INVALID_EVENT(event)) {
    return;
  }

  ESInternal *es = get_event_system_internal(event_system);
  if (event.id > es->event_count) {
    return;
  }

  EventSlot *event_slots_array = get_event_slots_array(es);
  u64 *free_array              = get_free_array(es);

  event_slots_array[event.id].registered = false;
  free_array[(es->free_count)++]         = event.id;

  if (event.id == es->event_count) {
    es->event_count -= 1;
  }
}

EventListener es_add_event_listener(EventSystem event_system, Event event, EventCallback callback) {
  EventListener listener = INVALID_LISTENER;

  if (IS_INVALID_SYSTEM(event_system) || IS_INVALID_EVENT(event) || !callback) {
    goto RETURN_LISTENER;
  }

  ESInternal *es    = get_event_system_internal(event_system);
  EventInternal *ev = get_event(es, event);

  if (ev->free_count > 0) {
    u64 id = ev->free[--(ev->free_count)];

    ev->free[ev->free_count] = 0;
    ev->callbacks[id]        = callback;

    listener.id    = id;
    listener.event = event;

    goto RETURN_LISTENER;
  }

  if (ev->count >= ev->capacity) {
    // TODO (Abdelrahman): Handle reallocating when out of callbacks
    goto RETURN_LISTENER;
  }

  u64 id = ++(ev->count);

  ev->callbacks[id] = callback;

  listener.id    = id;
  listener.event = event;

RETURN_LISTENER:
  return listener;
}

void es_remove_event_listener(EventSystem event_system, EventListener listener) {
  if (IS_INVALID_SYSTEM(event_system) || IS_INVALID_LISTENER(listener)) {
    return;
  }

  ESInternal *es = get_event_system_internal(event_system);
  if (listener.event.id > es->event_count) {
    return;
  }

  EventInternal *ev = get_event(es, listener.event);

  if (listener.id > ev->count) {
    return;
  }

  if (ev->callbacks[listener.id] == NULL) {
    return;
  }

  ev->callbacks[listener.id]   = NULL;
  ev->free[(ev->free_count)++] = listener.id; 

  if (listener.id == ev->count) {
    ev->count -= 1;
  }
}

void es_emit_event(EventSystem event_system, Event event, void *data) {
  if (IS_INVALID_SYSTEM(event_system) || IS_INVALID_EVENT(event)) {
    return;
  }

  ESInternal *es = get_event_system_internal(event_system);
  if (event.id > es->event_count) {
    return;
  }

  EventInternal *ev = get_event(es, event);
  EventCallback callback;
  for (int i = 0; i < ev->count; ++i) {
    callback = ev->callbacks[i + 1];
    if (!callback) {
      continue;
    }
    callback(data);
  }
}

// INTERNAL

internal void *alloc(u64 size) {
  return mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE | MAP_NORESERVE, -1, 0);
}

internal ESInternal *get_event_system_internal(EventSystem event_system) {
  return (ESInternal *)event_system.id;
}

internal EventSlot *get_event_slots_array(const ESInternal *es) {
  return (EventSlot *)((uptr)es + es->events_offset);
}

internal u64 *get_free_array(const ESInternal *es) {
  return (u64 *)((uptr)es + es->free_offset);
}

internal EventInternal *get_event(const ESInternal *es, Event event) {
  EventSlot *event_slots_array = get_event_slots_array(es);
  return get_event_from_offset(es, event_slots_array[event.id].offset);
}

internal EventInternal *get_event_from_offset(const ESInternal *es, u64 offset) {
  return (EventInternal *)((uptr)es + offset);
}
